# -*- coding: utf-8 -*-
"""The export command."""
import json

import click

from .utils import (get_datastore_api, partition_replace, save, execute_tasks)


@click.command()
@click.option('--project', '-p',
              help='GCP project.',
              required=True)
@click.option('--namespace', '-n',
              help='Datastore namespace.',
              required=True)
@click.option('--data-dir',
              help='Directory to be used to store exported files.',
              required=True,
              default='./data',
              show_default=True)
@click.option('--project-placeholder', '-pp',
              help='Placeholder value to replace the project value'
                   ' in exported files.',
              required=True,
              default='___PROJECT___',
              show_default=True)
@click.option('--namespace-placeholder', '-np',
              help='Placeholder value to replace the namespace value'
                   ' in exported files.',
              required=True,
              default='___NAMESPACE___',
              show_default=True)
@click.option('--kinds', '-k',
              help='Comma separated list of Datastore Kinds to export.',
              required=True)
def export(project, namespace, data_dir, project_placeholder,
           namespace_placeholder, kinds):
    """Export data from database."""
    execute_tasks({
        'type_task': 'export',
        'project': project,
        'namespace': namespace,
        'data_dir': data_dir,
        'project_placeholder': project_placeholder,
        'namespace_placeholder': namespace_placeholder,
        'kinds': kinds,
        'chunk': None,
        'target': execute_export
    })


def execute_export(project, namespace, data_dir, project_placeholder,
                   namespace_placeholder, kind, chunk):
    datastore = get_datastore_api()

    request_body = {
        'partitionId': {
            'projectId': project,
            'namespaceId': namespace
        },
        'query': {
            'kind': [
                {
                    'name': kind
                }
            ]
        }
    }

    entities_replaced = []
    more = True
    
    while more:
        response = datastore.projects() \
            .runQuery(projectId=project, body=request_body) \
            .execute()

        more = not (response['batch']['moreResults'] in ['NO_MORE_RESULTS'])
        cursor = response['batch']['endCursor']

        if more and cursor:
            request_body['query']['startCursor'] = cursor

        entities = extract_entities(response)
        entities_json = json.dumps(entities)
        entities_replaced_json = partition_replace(entities_json, project,
                                                project_placeholder, namespace,
                                                namespace_placeholder)
    
        entities_replaced.extend(json.loads(entities_replaced_json))

    save(entities_replaced, kind, data_dir)


def extract_entities(response):
    return [
        entityResult.get('entity')
        for entityResult in response.get('batch', {}).get('entityResults', [])
    ]



# -*- coding: utf-8 -*-
import click

from __version__ import __version__ as version
from commands.export import export
from commands.import_cmd import import_cmd


@click.group()
@click.version_option(version=version)
def main():
    """Utilities for Google Cloud Datastore."""
    pass


main.add_command(export)
main.add_command(import_cmd)



# -*- coding: utf-8 -*-
"""The import command."""
import json

import click

from .utils import (get_datastore_api, partition_replace, load, execute_tasks, split_lists)


@click.command('import')
@click.option('--project', '-p',
              help='GCP project.',
              required=True)
@click.option('--namespace', '-n',
              help='Datastore namespace.',
              required=True)
@click.option('--data-dir',
              help='Directory to be used to store exported files.',
              required=True,
              default='./data',
              show_default=True)
@click.option('--project-placeholder', '-pp',
              help='Placeholder value to replace the project value in'
                   ' previously exported files.',
              required=True,
              default='___PROJECT___',
              show_default=True)
@click.option('--namespace-placeholder', '-np',
              help='Placeholder value to replace the namespace value in'
                   ' previously exported files.',
              required=True,
              default='___NAMESPACE___',
              show_default=True)
@click.option('--kinds', '-k',
              help='Comma separated list of Datastore Kinds to import.',
              required=True)
@click.option('--chunk', '-c',
              help='A valid int number',
              default=500,
              required=False)
def import_cmd(project, namespace, data_dir, project_placeholder,
               namespace_placeholder, kinds, chunk):
    """Import data to database using previously exported data as input."""
    execute_tasks({
        'type_task': 'import',
        'project': project,
        'namespace': namespace,
        'data_dir': data_dir,
        'project_placeholder': project_placeholder,
        'namespace_placeholder': namespace_placeholder,
        'kinds': kinds,
        'chunk': chunk,
        'target': execute_import,
    })


def execute_import(project, namespace, data_dir, project_placeholder,
                   namespace_placeholder, kind, chunk):
    datastore = get_datastore_api()

    entities = load(kind, data_dir)
    entities_json = json.dumps(entities)
    entities_replaced_json = partition_replace(entities_json,
                                               project_placeholder, project,
                                               namespace_placeholder,
                                               namespace)
    entities_replaced = json.loads(entities_replaced_json)

    
    _iter = split_lists(entities_replaced, chunk)
    
    for _next in _iter:
        inserts = [
            {'insert': entity} for entity in _next
        ]

        request_body = {
            "mutations": inserts,
            "mode": "NON_TRANSACTIONAL"
        }

        datastore.projects() \
            .commit(projectId=project, body=request_body) \
            .execute()



# -*- coding: utf-8 -*-
"""Utilities."""
import copy
import io
import json
import os
import threading
from collections import namedtuple

import click
import googleapiclient.discovery

cls_task = namedtuple(
    'Task',
    [
        'type_task',
        'project',
        'namespace',
        'target',
        'data_dir',
        'project_placeholder',
        'namespace_placeholder',
        'kinds',
        'chunk'
    ]
)


def get_datastore_api():
    return googleapiclient.discovery.build('datastore', 'v1')


def get_kinds_list(kinds):
    return kinds.split(',')


def show_progressbar_item(item):
    if item is None:
        return 'Done!'
    return "Kind: {}".format(item)


def partition_replace(entities_json, from_project, to_project, from_namespace,
                      to_namespace):
    result = entities_json.replace('"projectId": "{}"'.format(from_project),
                                   '"projectId": "{}"'.format(to_project))
    result = result.replace('"namespaceId": "{}"'.format(from_namespace),
                            '"namespaceId": "{}"'.format(to_namespace))
    return result


def save(entities, kind, data_dir):
    if not entities:
        click.echo('No entities found for kind {}'.format(kind))
        return
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    with io.open('{}/{}.json'.format(data_dir, kind), 'w',
                 encoding='utf-8') as export_file:
        export_file.write(
            json.dumps(entities, ensure_ascii=False, sort_keys=True, indent=2,
                       separators=(u',', u': ')))


def load(kind, data_dir):
    with io.open('{}/{}.json'.format(data_dir, kind), 'r',
                 encoding='utf-8') as export_file:
        entities = json.load(export_file)
    return entities


def execute_tasks(kargs):
    data = cls_task(**kargs)
    kinds_list = get_kinds_list(data.kinds)
    click.echo(
        "Executing {}. Project={}, Namespace={}, Kinds={}.".format(
            data.type_task,
            data.project,
            data.namespace,
            kinds_list))

    tasks = {}
    for kind in kinds_list:
        tasks[kind] = threading.Thread(
            target=data.target,
            name=kind,
            args=(
                data.project, data.namespace, data.data_dir,
                data.project_placeholder,
                data.namespace_placeholder, kind, data.chunk
            )
        )

    kinds_list_progress = copy.deepcopy(kinds_list)

    click.echo('Starting tasks...')
    for task in tasks:
        tasks[task].start()

    click.echo('Done. {} tasks started.'.format(len(tasks)))
    click.echo('Executing...')

    while kinds_list_progress:
        for idx, kind in enumerate(kinds_list_progress):
            if not tasks.get(kind).is_alive():
                click.echo('Done. Kind={}'.format(kind))
                kinds_list_progress.pop(idx)

    click.echo('Finished!')

    
def split_lists(iterable, chunk_size):
    """Generate sequences of `chunk_size` elements from `iterable`."""
    iterable = iter(iterable)
    while True:
        chunk = []
        try:
            for _ in range(chunk_size):
                chunk.append(iterable.next())
            yield chunk
        except StopIteration:
            if chunk:
                yield chunk
            break